import time
from functools import cached_property, wraps

import cv2
import requests

from mower import __rootdir__
from mower.utils import config
from mower.utils.csleep import MowerExit
from mower.utils.device.method.adb import ADB
from mower.utils.image import bytes2img
from mower.utils.log import logger
from mower.utils.network import get_new_port, is_port_in_use


def retry_droidcast(func):
    @wraps(func)
    def retry_wrapper(self, *args, **kwargs):
        for _ in range(2):
            try:
                return func(self, *args, **kwargs)
            except MowerExit:
                raise
            except Exception as e:
                logger.exception(e)
                self.start_droidcast()

    return retry_wrapper


class DroidCast:
    def __init__(self):
        self.session = requests.Session()
        self.port = 0
        self.process = None

    @cached_property
    def adb(self) -> ADB:
        return config.device.get_service("adb")

    def get_droidcast_classpath(self) -> str | None:
        # TODO: 退出时（并非结束mower线程时）关闭DroidCast进程、取消ADB转发
        out = self.adb.adb_shell("pm path com.rayworks.droidcast")
        if out is None:
            logger.exception("无法获取CLASSPATH")
            return None
        prefix = "package:"
        postfix = ".apk"
        beg = out.index(prefix, 0)
        end = out.rfind(postfix)
        class_path = out[beg + len(prefix) : (end + len(postfix))].strip()
        class_path = "CLASSPATH=" + class_path
        logger.info(f"成功获取CLASSPATH：{class_path}")
        return class_path

    def start_droidcast(self) -> bool:
        class_path = self.get_droidcast_classpath()
        if not class_path:
            logger.info("安装DroidCast")
            apk_path = f"{__rootdir__}/vendor/DroidCast-debug-1.2.1.apk"
            try:
                self.adb.adb.install(apk_path)
                logger.info("DroidCast安装完成，获取CLASSPATH")
            except Exception as e:
                logger.error(f"DroidCast安装失败: {e}")
                return False
            class_path = self.get_droidcast_classpath()
            if not class_path:
                logger.error("无法获取CLASSPATH")
                return False
        port = self.port
        if port != 0 and is_port_in_use(port):
            try:
                forward_list = self.adb.adb.forward_list()
                for forward in forward_list:
                    if (
                        forward.local == f"tcp:{port}"
                        and forward.remote == f"tcp:{port}"
                    ):
                        break
            except Exception as e:
                logger.exception(e)
        if port == 0:
            port = get_new_port()
            self.port = port
            logger.info(f"更新DroidCast端口为{port}")
        else:
            logger.info(f"保持DroidCast端口为{port}")
        self.adb.adb.forward(f"tcp:{port}", f"tcp:{port}")
        logger.info("ADB端口转发成功，启动DroidCast")
        self.process = self.adb.process(
            class_path,
            [
                "app_process",
                "/",
                "com.rayworks.droidcast.Main",
                f"--port={port}",
            ],
        )
        time.sleep(1)  # TODO: 更好地等待DroidCast启动
        return True

    @retry_droidcast
    def capture_display(self):
        try:
            url = f"http://127.0.0.1:{self.port}/screenshot"
            logger.debug(f"GET {url}")
            r = self.session.get(url)
            img = bytes2img(r.content)
            if config.conf.droidcast.rotate:
                img = cv2.rotate(img, cv2.ROTATE_180)
            return img
        except Exception as e:
            raise e
